import os

from brightsky.cli import cli
from brightsky.utils import configure_logging, load_dotenv


if __name__ == '__main__':
    load_dotenv()
    configure_logging()
    if os.getenv('SENTRY_DSN'):
        import sentry_sdk
        sample_rate = os.getenv('SENTRY_TRACES_SAMPLE_RATE')
        if sample_rate:
            sample_rate = float(sample_rate)
        sentry_sdk.init(
            dsn=os.getenv('SENTRY_DSN'),
            traces_sample_rate=sample_rate)
    cli(prog_name='python -m brightsky')



import datetime

import falcon
import pytest
from dateutil.tz import tzutc

import brightsky
from brightsky.export import DBExporter, SYNOPExporter
from brightsky.web import make_app

from .utils import settings


SOURCES = [
    # Ordered by their distance to (52, 7.6)
    {
        'observation_type': 'forecast',
        'lat': 51.97,
        'lon': 7.63,
        'height': 60,
        'station_name': 'MUENSTER ZENTRUM',
        'dwd_station_id': None,
        'wmo_station_id': 'P0036',
    },
    {
        'observation_type': 'current',
        'lat': 52.13,
        'lon': 7.7,
        'height': 48,
        'station_name': 'MUENSTER/OSNABR.',
        'dwd_station_id': '01766',
        'wmo_station_id': '10315',
    },
    {
        'observation_type': 'historical',
        'lat': 52.1344,
        'lon': 7.6969,
        'height': 47.8,
        'station_name': 'Münster/Osnabrück',
        'dwd_station_id': '01766',
        'wmo_station_id': '10315',
    },
]
RECENT_RECORDS = [
    {
        'timestamp': (
            datetime.datetime(2020, 8, 19, 12, tzinfo=tzutc())
            + datetime.timedelta(hours=i)),
        **SOURCES[2],
        'temperature': 100 + i,
    }
    for i in range(40)
]
CURRENT_RECORDS = [
    {
        'timestamp': (
            datetime.datetime(2020, 8, 20, 12, tzinfo=tzutc())
            + datetime.timedelta(hours=i)),
        **SOURCES[1],
        'temperature': 200 + i,
    }
    for i in range(24)
]
FORECAST_RECORDS = [
    {
        'timestamp': (
            datetime.datetime(2020, 8, 21, 8, tzinfo=tzutc())
            + datetime.timedelta(hours=i)),
        **SOURCES[0],
        'temperature': 300 + i,
    }
    for i in range(40)
]

ALL_FIELDS_RECORD = {
    'temperature': 294.35,
    'precipitation': 1.4,
    'pressure_msl': 100620,
    'sunshine': 1740,
    'wind_direction': 130,
    'wind_speed': 2.2,
    'cloud_cover': 87,
    'dew_point': 287.65,
    'visibility': 40000,
    'wind_gust_direction': 130,
    'wind_gust_speed': 3.9,
    'condition': 'fog',
}
RECENT_RECORDS[12].update(ALL_FIELDS_RECORD)

CONDITION_FIELDS = [
    {'condition': 'fog', '_expected_icon': 'fog'},
    {'condition': 'dry', '_expected_icon': 'clear-night'},
    {'condition': 'dry', 'cloud_cover': 50.,
     '_expected_icon': 'partly-cloudy-night'},
    {'condition': 'sleet', '_expected_icon': 'sleet'},
    {'condition': 'snow', '_expected_icon': 'snow'},
    {'condition': 'hail', '_expected_icon': 'hail'},
    {'condition': 'thunderstorm', '_expected_icon': 'thunderstorm'},
    {'condition': 'dry', '_expected_icon': 'clear-day'},
    {'condition': 'rain', '_expected_icon': 'rain'},
    {'condition': 'dry', 'cloud_cover': 50.,
     '_expected_icon': 'partly-cloudy-day'},
    {'condition': 'dry', 'cloud_cover': 100., '_expected_icon': 'cloudy'},
    {'precipitation': 0.3, '_expected_icon': 'clear-day'},
    {'precipitation': 10., '_expected_icon': 'rain'},
]
for i, fields in enumerate(CONDITION_FIELDS):
    RECENT_RECORDS[12+i].update(fields)


SYNOP_SOURCE = {
    'observation_type': 'synop',
    'lat': 52.1344,
    'lon': 7.69685,
    'height': 47.8,
    'station_name': 'Muenster/Osnabrueck',
    'dwd_station_id': '01766',
    'wmo_station_id': '10315',
}
SYNOP_NOW = datetime.datetime.utcnow().replace(
    minute=0, second=0, microsecond=0, tzinfo=tzutc()
) + datetime.timedelta(hours=1)
SYNOP_RECORDS = [
    {
        'timestamp': SYNOP_NOW - datetime.timedelta(minutes=90),
        **SYNOP_SOURCE,
        'temperature': 300,
    },
    {
        'timestamp': SYNOP_NOW - datetime.timedelta(minutes=50),
        **SYNOP_SOURCE,
        'precipitation_10': 0.2,
        'wind_speed_10': 2.0,
        'wind_direction_10': 10,
        'wind_gust_speed_10': 2.3,
        'wind_gust_direction_10': 130,
    },
    {
        'timestamp': SYNOP_NOW - datetime.timedelta(minutes=40),
        **SYNOP_SOURCE,
        'precipitation_10': 0.3,
        'wind_speed_10': 1.8,
        'wind_direction_10': 350,
        'wind_gust_speed_10': 5.7,
        'wind_gust_direction_10': 60,
    },
    {
        'timestamp': SYNOP_NOW - datetime.timedelta(minutes=30),
        **SYNOP_SOURCE,
        'cloud_cover': None,
        'dew_point': 292.17,
        'precipitation_10': 0.,
        'pressure_msl': 100880,
        'relative_humidity': 80,
        'sunshine_30': 1440,
        'temperature': 296.15,
        'visibility': 35000,
        'wind_direction_10': 340,
        'wind_speed_10': 1.2,
        'wind_gust_direction_10': 20,
        'wind_gust_speed_10': 2.1,
    },
    {
        'timestamp': SYNOP_NOW - datetime.timedelta(minutes=20),
        **SYNOP_SOURCE,
        'precipitation_10': 0.,
        'wind_speed_10': 0.8,
        'wind_direction_10': 30,
        'wind_gust_speed_10': 3.3,
        'wind_gust_direction_10': 30,
    },
    {
        'timestamp': SYNOP_NOW - datetime.timedelta(minutes=10),
        **SYNOP_SOURCE,
        'precipitation_10': 0.1,
        'wind_speed_10': 2.2,
        'wind_direction_10': 20,
        'wind_gust_speed_10': 5.1,
        'wind_gust_direction_10': 40,
    },
    {
        'timestamp': SYNOP_NOW,
        **SYNOP_SOURCE,
        'cloud_cover': 88,
        'dew_point': 292.47,
        'precipitation_10': 0,
        'pressure_msl': 100920,
        'relative_humidity': 80,
        'sunshine_60': 3000,
        'temperature': 296.05,
        'wind_direction_10': 350,
        'wind_speed_10': 2.1,
        'wind_gust_direction_10': 30,
        'wind_gust_speed_10': 3.1,
        'condition': 'rain',
    },
]


@pytest.fixture
def data(db):
    records = RECENT_RECORDS + CURRENT_RECORDS + FORECAST_RECORDS
    DBExporter().export(records)


@pytest.fixture
def synop_data(db):
    SYNOPExporter().export(SYNOP_RECORDS)


def test_sources_required_parameters(data, api):
    assert api.simulate_get('/sources').status_code == 400
    assert api.simulate_get('/sources?lat=52').status_code == 400
    assert api.simulate_get('/sources?lon=7.6').status_code == 400
    assert api.simulate_get('/sources?lat=52&lon=7.6').status_code == 200
    assert api.simulate_get('/sources?wmo_station_id=10315').status_code == 200
    assert api.simulate_get('/sources?dwd_station_id=01766').status_code == 200


def test_sources_response(data, api):
    resp = api.simulate_get('/sources?lat=52&lon=7.6')
    assert resp.status_code == 200
    resp_sources = resp.json['sources']
    assert len(resp_sources) == 3
    for resp_source, source in zip(resp_sources, SOURCES):
        for k, v in source.items():
            assert resp_source[k] == v
    assert [s['distance'] for s in resp.json['sources']] == [
        3922, 16008, 16365]
    assert resp_sources[0]['first_record'] == (
        FORECAST_RECORDS[0]['timestamp'].isoformat())
    assert resp_sources[0]['last_record'] == (
        FORECAST_RECORDS[-1]['timestamp'].isoformat())


def test_sources_max_dist(data, api):
    resp = api.simulate_get('/sources?lat=52&lon=7.6')
    assert len(resp.json['sources']) == 3
    resp = api.simulate_get('/sources?lat=52&lon=7.6&max_dist=5000')
    assert len(resp.json['sources']) == 1


def test_sources_by_station_id(data, api):
    paths = [
        '/sources?wmo_station_id=10315',
        '/sources?dwd_station_id=01766',
    ]
    for path in paths:
        resp = api.simulate_get(path)
        assert resp.status_code == 200
        assert len(resp.json['sources']) == 2


def test_source_by_source_id(db, data, api):
    source_id = db.fetch(
        "SELECT id FROM sources WHERE observation_type = 'forecast'")[0]['id']
    resp = api.simulate_get(f'/sources?source_id={source_id}')
    assert len(resp.json['sources']) == 1
    assert resp.json['sources'][0]['observation_type'] == 'forecast'


def test_no_sources_available(data, api):
    assert api.simulate_get('/sources?lat=0&lon=0').status_code == 404
    assert api.simulate_get('/sources?wmo_station_id=12345').status_code == 404
    assert api.simulate_get('/sources?dwd_station_id=12345').status_code == 404


def test_weather_required_parameters(data, api):
    assert api.simulate_get('/weather').status_code == 400
    assert api.simulate_get('/weather?lat=52&lon=7.6').status_code == 400
    assert api.simulate_get(
        '/weather?lat=52&lon=7.6&date=2020-08-20').status_code == 200
    assert api.simulate_get(
        '/weather?lat=52&lon=7.6&last_date=2020-08-20').status_code == 400


def test_weather_response(data, api):
    resp = api.simulate_get('/weather?lat=52&lon=7.6&date=2020-08-20')
    assert len(resp.json['weather']) == 25
    assert len(resp.json['sources']) == 1
    assert all(
        w['source_id'] == resp.json['sources'][0]['id']
        for w in resp.json['weather'])
    assert resp.json['weather'][0]['timestamp'] == '2020-08-20T00:00:00+00:00'
    assert resp.json['weather'][-1]['timestamp'] == '2020-08-21T00:00:00+00:00'
    for w, record in zip(resp.json['weather'], RECENT_RECORDS[12:]):
        assert w['temperature'] == round(record['temperature'] - 273.15, 2)


def test_weather_date_range(data, api):
    resp = api.simulate_get(
        '/weather?lat=52&lon=7.6&date=2020-08-20&last_date=2020-08-22')
    assert len(resp.json['weather']) == 49
    assert resp.json['weather'][0]['timestamp'] == '2020-08-20T00:00:00+00:00'
    assert resp.json['weather'][-1]['timestamp'] == '2020-08-22T00:00:00+00:00'
    resp = api.simulate_get('/weather?lat=52&lon=7.6&date=2020-08-20T12:00')
    assert len(resp.json['weather']) == 25
    assert resp.json['weather'][0]['timestamp'] == '2020-08-20T12:00:00+00:00'
    assert resp.json['weather'][-1]['timestamp'] == '2020-08-21T12:00:00+00:00'


def test_weather_source_selection(data, api):
    resp = api.simulate_get(
        '/weather?lat=52&lon=7.6&date=2020-08-20&last_date=2020-08-22')
    assert len(resp.json['sources']) == 3
    observation_types = {
        s['id']: s['observation_type'] for s in resp.json['sources']}
    for w in resp.json['weather'][:28]:
        assert observation_types[w['source_id']] == 'historical'
    for w in resp.json['weather'][28:36]:
        assert observation_types[w['source_id']] == 'current'
    for w in resp.json['weather'][36:]:
        assert observation_types[w['source_id']] == 'forecast'


def test_weather_units(data, api):
    resp_dwd = api.simulate_get('/weather?lat=52&lon=7.6&date=2020-08-20')
    resp_si = api.simulate_get(
        '/weather?lat=52&lon=7.6&date=2020-08-20&units=si')
    expected_conversions = {
        'temperature': 21.2,
        'pressure_msl': 1006.2,
        'sunshine': 29,
        'wind_speed': 7.9,
        'dew_point': 14.5,
        'wind_gust_speed': 14.,
    }
    for k, v in ALL_FIELDS_RECORD.items():
        assert resp_dwd.json['weather'][0][k] == expected_conversions.get(k, v)
        assert resp_si.json['weather'][0][k] == v


def test_weather_timezone(data, api):
    resp = api.simulate_get(
        '/weather?lat=52&lon=7.6&date=2020-08-20&tz=Europe/Berlin')
    assert resp.json['weather'][0]['timestamp'] == '2020-08-20T00:00:00+02:00'
    assert resp.json['weather'][-1]['timestamp'] == '2020-08-21T00:00:00+02:00'
    # date offset should be used if tz not supplied
    resp = api.simulate_get(
        '/weather?lat=52&lon=7.6&date=2020-08-20T00:00:00%2b02:00')
    assert resp.json['weather'][0]['timestamp'] == '2020-08-20T00:00:00+02:00'
    assert resp.json['weather'][-1]['timestamp'] == '2020-08-21T00:00:00+02:00'
    # but tz should always be preferred
    resp = api.simulate_get(
        '/weather?lat=52&lon=7.6&date=2020-08-20T00:00:00-04:00'
        '&tz=Europe/Berlin')
    assert resp.json['weather'][0]['timestamp'] == '2020-08-20T06:00:00+02:00'
    assert resp.json['weather'][-1]['timestamp'] == '2020-08-21T06:00:00+02:00'
    # last_date's offset should be used for range selection but not as response
    # timezone
    resp = api.simulate_get(
        '/weather?lat=52&lon=7.6&date=2020-08-20'
        '&last_date=2020-08-20T12:00-04:00')
    assert resp.json['weather'][0]['timestamp'] == '2020-08-20T00:00:00+00:00'
    assert resp.json['weather'][-1]['timestamp'] == '2020-08-20T16:00:00+00:00'
    # date's offset should NOT be used as default last_date offset
    resp = api.simulate_get(
        '/weather?lat=52&lon=7.6&date=2020-08-20T00:00:00%2b02:00'
        '&last_date=2020-08-22')
    assert resp.json['weather'][0]['timestamp'] == '2020-08-20T00:00:00+02:00'
    assert resp.json['weather'][-1]['timestamp'] == '2020-08-22T02:00:00+02:00'


def test_weather_icon(data, api):
    resp = api.simulate_get('/weather?lat=52&lon=7.6&date=2020-08-20')
    for condition, record in zip(CONDITION_FIELDS, resp.json['weather']):
        assert record['icon'] == condition['_expected_icon']


def test_synop_disallows_lat_lon(data, api):
    resp = api.simulate_get(
        '/synop',
        params={
            'lat': 52,
            'lon': 7.6,
            'date': SYNOP_RECORDS[0]['timestamp'].isoformat(),
        })
    assert resp.status_code == 400


def test_synop_response(synop_data, api):
    resp = api.simulate_get(
        '/synop',
        params={
            'wmo_station_id': SYNOP_SOURCE['wmo_station_id'],
            'date': SYNOP_RECORDS[0]['timestamp'].isoformat(),
        })
    assert resp.status_code == 200
    assert len(resp.json['sources']) == 1
    assert len(resp.json['weather']) == len(SYNOP_RECORDS)
    for w, record in zip(resp.json['weather'], SYNOP_RECORDS):
        assert w['timestamp'] == record['timestamp'].isoformat()


def test_current_weather_response(synop_data, api, db):
    # XXX: This test may be flaky as the concurrent refresh may not have
    #      finished yet. Can we somehow wait until the lock is released?
    resp = api.simulate_get('/current_weather?lat=52&lon=7.6')
    assert resp.status_code == 200
    assert len(resp.json['sources']) == 1
    expected_weather = {
        # From latest record
        'timestamp': SYNOP_NOW.isoformat(),
        'cloud_cover': 88,
        'condition': 'rain',
        'dew_point': 19.32,
        'precipitation_10': 0.,
        'pressure_msl': 1009.2,
        'relative_humidity': 80,
        'temperature': 22.9,
        'wind_direction_10': 350,
        'wind_speed_10': 7.6,
        'wind_gust_direction_10': 30,
        'wind_gust_speed_10': 11.2,
        # Summed from latest three/six records
        'precipitation_30': 0.1,
        'precipitation_60': 0.6,
        # Averaged from latest three/six records
        'wind_speed_30': 6.1,
        'wind_speed_60': 6.1,
        # Averaged (for circular quantities) from latest three/six records
        'wind_direction_30': 13,
        'wind_direction_60': 3,
        # Filled up from previous record
        'visibility': 35000,
        # Maximum from previous three/six records
        'wind_gust_direction_30': 40,
        'wind_gust_direction_60': 60,
        'wind_gust_speed_30': 18.4,
        'wind_gust_speed_60': 20.5,
        # Calculated from latest available _30 and _60 values
        'sunshine_30': 26,
        'sunshine_60': 50,
    }
    for k, v in expected_weather.items():
        assert resp.json['weather'][k] == v, k


def test_status_response(api):
    resp = api.simulate_get('/')
    assert resp.status_code == 200
    assert resp.json['name'] == 'brightsky'
    assert resp.json['version'] == brightsky.__version__
    assert resp.json['status'] == 'ok'


def test_cors(synop_data, db):
    def _get_response(**kwargs):
        api = falcon.testing.TestClient(make_app())
        return api.simulate_get('/current_weather?lat=52&lon=7.6', **kwargs)

    example_com = 'http://example.com'
    brightsky_dev = 'https://brightsky.dev'

    with settings(CORS_ALLOW_ALL_ORIGINS=True):
        resp = _get_response(headers={'Origin': example_com})
        assert resp.headers['access-control-allow-origin'] == '*'

    with settings(CORS_ALLOWED_ORIGINS=[example_com, brightsky_dev]):
        resp = _get_response(headers={'Origin': example_com})
        assert resp.headers['access-control-allow-origin'] == example_com
        resp = _get_response(headers={'Origin': brightsky_dev})
        assert resp.headers['access-control-allow-origin'] == brightsky_dev

    with settings(CORS_ALLOWED_ORIGINS=[brightsky_dev]):
        resp = _get_response(headers={'Origin': example_com})
        assert 'access-control-allow-origin' not in resp.headers
        resp = _get_response(headers={'Origin': brightsky_dev})
        assert resp.headers['access-control-allow-origin'] == brightsky_dev



import datetime

from dateutil.tz import tzutc

from brightsky.parsers import (
    CloudCoverObservationsParser, CurrentObservationsParser,
    DewPointObservationsParser, get_parser, MOSMIXParser,
    PrecipitationObservationsParser, PressureObservationsParser,
    SunshineObservationsParser, SYNOPParser, TemperatureObservationsParser,
    VisibilityObservationsParser, WindGustsObservationsParser,
    WindObservationsParser)

from .utils import is_subset, settings


def test_mosmix_parser(data_dir):
    p = MOSMIXParser()
    records = list(p.parse(data_dir / 'MOSMIX_S.kmz'))
    assert len(records) == 240
    assert records[0] == {
        'observation_type': 'forecast',
        'source': 'MOSMIX:2020-03-13T09:00:00.000Z',
        'lat': 74.52,
        'lon': 19.02,
        'height': 16.,
        'dwd_station_id': 'XXX',
        'wmo_station_id': '01028',
        'station_name': 'BJORNOYA',
        'timestamp': datetime.datetime(2020, 3, 13, 10, 0, tzinfo=tzutc()),
        'cloud_cover': 93.0,
        'dew_point': 257.25,
        'precipitation': 0.1,
        'precipitation_probability_6h': None,
        'pressure_msl': 99000.0,
        'sunshine': None,
        'temperature': 260.45,
        'visibility': 1700.0,
        'wind_direction': 330.0,
        'wind_speed': 8.75,
        'wind_gust_speed': None,
        'condition': 'snow',
        'solar': None,
    }
    assert records[-1] == {
        'observation_type': 'forecast',
        'source': 'MOSMIX:2020-03-13T09:00:00.000Z',
        'lat': 74.52,
        'lon': 19.02,
        'height': 16.,
        'dwd_station_id': 'XXX',
        'wmo_station_id': '01028',
        'station_name': 'BJORNOYA',
        'timestamp': datetime.datetime(2020, 3, 23, 9, 0, tzinfo=tzutc()),
        'cloud_cover': 76.,
        'dew_point': 265.35,
        'precipitation': None,
        'precipitation_probability_6h': None,
        'pressure_msl': 100630.0,
        'sunshine': None,
        'temperature': 267.15,
        'visibility': 11600.0,
        'wind_direction': 49.0,
        'wind_speed': 7.72,
        'wind_gust_speed': None,
        'condition': 'dry',
        'solar': None,
    }
    assert records[2]['precipitation_probability_6h'] == 49


def test_synop_parser(data_dir):
    p = SYNOPParser()
    records = list(p.parse(data_dir / 'synop.json.bz2'))
    assert len(records) == 3
    assert records[0] == {
        'observation_type': 'synop',
        'lat': 52.1344,
        'lon': 7.69685,
        'height': 47.8,
        'wmo_station_id': '10315',
        'dwd_station_id': '01766',
        'station_name': 'Muenster/Osnabrueck',
        'timestamp': datetime.datetime(2020, 6, 17, 9, 0, tzinfo=tzutc()),
        'cloud_cover': 88,
        'dew_point': 287.37,
        'pressure_msl': 101290,
        'relative_humidity': 66,
        'temperature': 294.05,
        'wind_direction_10': 30,
        'wind_speed_10': 2,
        'wind_gust_direction_10': None,
        'wind_gust_speed_10': None,
        'condition': 'dry',
    }
    assert records[1] == {
        'observation_type': 'synop',
        'lat': 52.1344,
        'lon': 7.69685,
        'height': 47.8,
        'wmo_station_id': '10315',
        'dwd_station_id': '01766',
        'station_name': 'Muenster/Osnabrueck',
        'timestamp': datetime.datetime(2020, 6, 17, 9, 0, tzinfo=tzutc()),
        'visibility': None,
        'sunshine_60': 2520,
        'precipitation_60': 0,
        'wind_direction_10': None,
        'wind_speed_10': None,
        'wind_gust_direction_10': None,
        'wind_gust_speed_10': None,
        'wind_gust_direction_60': None,
        'wind_gust_speed_60': 4.6,
        'wind_gust_direction_30': 340,
        'wind_gust_speed_30': 4.6,
    }
    assert records[2]['wmo_station_id'] == 'M031'
    assert records[2]['dwd_station_id'] == '05484'


def test_current_observation_parser(data_dir):
    p = CurrentObservationsParser()
    path = data_dir / 'observations_current.csv'
    records = list(p.parse(path, 10.1, 20.2, 30.3, 'Muenster'))
    assert len(records) == 25
    assert records[0] == {
        'observation_type': 'current',
        'lat': 10.1,
        'lon': 20.2,
        'height': 30.3,
        'dwd_station_id': 'YYY',
        'wmo_station_id': '01049',
        'station_name': 'Muenster',
        'timestamp': datetime.datetime(2020, 4, 6, 8, 0, tzinfo=tzutc()),
        'cloud_cover': None,
        'dew_point': 263.15,
        'precipitation': 0,
        'pressure_msl': 102310.,
        'relative_humidity': 59.,
        'sunshine': None,
        'temperature': 270.05,
        'visibility': None,
        'wind_direction': 140,
        'wind_speed': 3.9,
        'wind_gust_speed': 5.8,
        'condition': None,
        'solar': None,
    }
    assert records[15] == {
        'observation_type': 'current',
        'lat': 10.1,
        'lon': 20.2,
        'height': 30.3,
        'dwd_station_id': 'YYY',
        'wmo_station_id': '01049',
        'station_name': 'Muenster',
        'timestamp': datetime.datetime(2020, 4, 5, 17, 0, tzinfo=tzutc()),
        'cloud_cover': None,
        'dew_point': 270.05,
        'precipitation': 0.6,
        'pressure_msl': 101910.,
        'relative_humidity': 94.,
        'sunshine': None,
        'temperature': 270.95,
        'visibility': None,
        'wind_direction': 230,
        'wind_speed': 3.9,
        'wind_gust_speed': 7.2,
        'condition': None,
        'solar': None,
    }


def test_current_observation_parser_loads_station_data_from_db(db, data_dir):
    source = {
        'observation_type': 'forecast',
        'lat': 52.12,
        'lon': 7.62,
        'height': 5.,
        'wmo_station_id': '01049',
        'station_name': 'Münster',
    }
    db.insert('sources', [source])
    p = CurrentObservationsParser()
    record = next(p.parse(data_dir / 'observations_current.csv'))
    for field in ('lat', 'lon', 'height', 'station_name'):
        assert record[field] == source[field]


def test_observations_parser_parses_metadata(data_dir):
    p = WindObservationsParser()
    metadata = {
        'observation_type': 'historical',
        'source': (
            'Observations:Recent:produkt_ff_stunde_20180915_20200317_04911.txt'
        ),
        'lat': 48.8275,
        'lon': 12.5597,
        'height': 350.5,
        'dwd_station_id': '04911',
        'wmo_station_id': '10788',
        'station_name': 'Straubing',
    }
    for record in p.parse(data_dir / 'observations_recent_FF_akt.zip'):
        assert is_subset(metadata, record)


def test_observations_parser_handles_missing_values(data_dir):
    p = WindObservationsParser()
    records = list(p.parse(data_dir / 'observations_recent_FF_akt.zip'))
    assert records[5]['wind_direction'] == 90
    assert records[5]['wind_speed'] is None


def test_observations_parser_handles_ignored_values(data_dir):
    p = WindObservationsParser()
    p.ignored_values = {'wind_direction': ['80']}
    records = list(p.parse(path=data_dir / 'observations_recent_FF_akt.zip'))
    assert records[0]['wind_direction'] is None
    assert records[0]['wind_speed'] == 1.6


def test_observations_parser_handles_location_changes(data_dir):
    p = WindObservationsParser()
    path = data_dir / 'observations_recent_FF_location_change_akt.zip'
    records = list(p.parse(path))
    assert is_subset(
        {'lat': 48.8275, 'lon': 12.5597, 'height': 350.5}, records[0])
    assert is_subset(
        {'lat': 50.0, 'lon': 13.0, 'height': 345.0}, records[-1])


def test_observations_parser_skips_file_if_out_of_range(data_dir):
    p = PressureObservationsParser()
    path = data_dir / 'observations_19950901_20150817_hist.zip'
    assert not p.skip_path(path)
    with settings(
        MIN_DATE=datetime.datetime(2016, 1, 1, tzinfo=tzutc()),
    ):
        assert p.skip_path(path)
    with settings(
        MAX_DATE=datetime.datetime(1995, 1, 1, tzinfo=tzutc()),
    ):
        assert p.skip_path(path)


def test_observations_parser_skips_rows_if_before_cutoff(data_dir):
    p = WindObservationsParser()
    path = data_dir / 'observations_recent_FF_akt.zip'
    with settings(
        MIN_DATE=datetime.datetime(2019, 1, 1, tzinfo=tzutc()),
    ):
        records = list(p.parse(path))
        assert len(records) == 5
        assert records[0]['timestamp'] == datetime.datetime(
            2019, 4, 20, 21, tzinfo=tzutc())
    with settings(
        MAX_DATE=datetime.datetime(2019, 1, 1, tzinfo=tzutc()),
    ):
        records = list(p.parse(path))
        assert len(records) == 5
        assert records[-1]['timestamp'] == datetime.datetime(
            2018, 9, 15, 4, tzinfo=tzutc())


def _test_parser(
        cls, path, first, last, count=10, first_idx=0, last_idx=-1, **kwargs):
    records = list(cls().parse(path, **kwargs))
    first['timestamp'] = datetime.datetime.strptime(
        first['timestamp'], '%Y-%m-%d %H:%M').replace(tzinfo=tzutc())
    last['timestamp'] = datetime.datetime.strptime(
        last['timestamp'], '%Y-%m-%d %H:%M').replace(tzinfo=tzutc())
    assert len(records) == count
    assert is_subset(first, records[first_idx])
    assert is_subset(last, records[last_idx])


def test_cloud_cover_observations_parser(data_dir):
    _test_parser(
        CloudCoverObservationsParser,
        data_dir / 'observations_recent_N_akt.zip',
        {'timestamp': '2018-12-03 07:00', 'cloud_cover': 50},
        {'timestamp': '2019-11-20 00:00', 'cloud_cover': None},
    )


def test_dew_point_observations_parser(data_dir):
    _test_parser(
        DewPointObservationsParser,
        data_dir / 'observations_recent_TD_akt.zip',
        {'timestamp': '2018-12-03 00:00', 'dew_point': 284.55},
        {'timestamp': '2020-05-29 15:00', 'dew_point': 271.65},
    )


def test_temperature_observations_parser(data_dir):
    _test_parser(
        TemperatureObservationsParser,
        data_dir / 'observations_recent_TU_akt.zip',
        {'timestamp': '2018-09-15 00:00',
         'temperature': 286.85, 'relative_humidity': 96},
        {'timestamp': '2020-03-17 23:00',
         'temperature': 275.75, 'relative_humidity': 100},
    )


def test_precipitation_observations_parser(data_dir):
    _test_parser(
        PrecipitationObservationsParser,
        data_dir / 'observations_recent_RR_akt.zip',
        {'timestamp': '2018-09-22 20:00', 'precipitation': 0.0},
        {'timestamp': '2020-02-11 02:00', 'precipitation': 0.3},
    )


def test_visibility_observations_parser(data_dir):
    _test_parser(
        VisibilityObservationsParser,
        data_dir / 'observations_recent_VV_akt.zip',
        {'timestamp': '2018-12-03 00:00', 'visibility': 15000},
        {'timestamp': '2020-06-04 23:00', 'visibility': 30000},
    )


def test_wind_observations_parser(data_dir):
    _test_parser(
        WindObservationsParser,
        data_dir / 'observations_recent_FF_akt.zip',
        {'timestamp': '2018-09-15 00:00',
         'wind_speed': 1.6, 'wind_direction': 80},
        {'timestamp': '2020-03-17 23:00',
         'wind_speed': 1.5, 'wind_direction': 130},
    )


def test_wind_gusts_observations_parser(data_dir):
    _test_parser(
        WindGustsObservationsParser,
        data_dir / 'observations_recent_extrema_wind_akt.zip',
        {'timestamp': '2018-12-03 00:00',
         'wind_gust_speed': 6.3, 'wind_gust_direction': 210},
        {'timestamp': '2020-06-04 23:00',
         'wind_gust_speed': 6.2, 'wind_gust_direction': 270},
        meta_path=data_dir / 'observations_recent_extrema_wind_akt_meta.zip'
    )


def test_sunshine_observations_parser(data_dir):
    _test_parser(
        SunshineObservationsParser,
        data_dir / 'observations_recent_SD_akt.zip',
        {'timestamp': '2018-09-15 11:00', 'sunshine': 600.},
        {'timestamp': '2020-03-17 16:00', 'sunshine': 0.},
        first_idx=2,
    )


def test_pressure_observations_parser(data_dir):
    _test_parser(
        PressureObservationsParser,
        data_dir / 'observations_recent_P0_hist.zip',
        {'timestamp': '2018-09-15 00:00', 'pressure_msl': 102120},
        {'timestamp': '2020-03-17 23:00', 'pressure_msl': 103190},
    )


def test_pressure_observations_parser_approximates_pressure_msl(data_dir):
    p = PressureObservationsParser()
    records = list(p.parse(data_dir / 'observations_recent_P0_hist.zip'))
    # The actual reduced pressure deleted from the test observation file was
    # 1023.0 hPa
    assert records[4]['pressure_msl'] == 102260


def test_get_parser():
    synop_with_timestamp = (
        'Z__C_EDZW_20200617114802_bda01,synop_bufr_GER_999999_999999__MW_617'
        '.json.bz2')
    synop_latest = (
        'Z__C_EDZW_latest_bda01,synop_bufr_GER_999999_999999__MW_XXX.json.bz2')
    expected = {
        '10minutenwerte_extrema_wind_00427_akt.zip': (
            WindGustsObservationsParser),
        'stundenwerte_FF_00011_akt.zip': WindObservationsParser,
        'stundenwerte_FF_00090_akt.zip': WindObservationsParser,
        'stundenwerte_N_01766_akt.zip': CloudCoverObservationsParser,
        'stundenwerte_P0_00096_akt.zip': PressureObservationsParser,
        'stundenwerte_RR_00102_akt.zip': PrecipitationObservationsParser,
        'stundenwerte_SD_00125_akt.zip': SunshineObservationsParser,
        'stundenwerte_TD_01766.zip': DewPointObservationsParser,
        'stundenwerte_TU_00161_akt.zip': TemperatureObservationsParser,
        'stundenwerte_VV_00161_akt.zip': VisibilityObservationsParser,
        'MOSMIX_S_LATEST_240.kmz': MOSMIXParser,
        'K611_-BEOB.csv': CurrentObservationsParser,
        synop_with_timestamp: SYNOPParser,
        synop_latest: None,
    }
    for filename, expected_parser in expected.items():
        assert get_parser(filename) is expected_parser



import logging
import os
import tempfile

from brightsky.db import get_connection
from brightsky.parsers import get_parser
from brightsky.polling import DWDPoller
from brightsky.utils import download
from brightsky.worker import huey, process


logger = logging.getLogger('brightsky')


def parse(url):
    parser = get_parser(os.path.basename(url))()
    with tempfile.TemporaryDirectory() as tmpdir:
        path, fingerprint = download(url, tmpdir)
        extra = {
            kwarg: download(extra_url, tmpdir)[0]
            for kwarg, extra_url in parser.get_extra_urls(path).items()
        }
        exporter = parser.exporter()
        exporter.export(parser.parse(path, **extra), fingerprint=fingerprint)


def poll(enqueue=False):
    updated_files = DWDPoller().poll()
    if enqueue:
        if (expired_locks := huey.expire_locks(1800)):
            logger.warning(
                'Removed expired locks: %s', ', '.join(expired_locks))
        pending_urls = [
            t.args[0] for t in huey.pending() if t.name == 'process']
        enqueued = 0
        for updated_file in updated_files:
            url = updated_file['url']
            if url in pending_urls:
                logger.debug('Skipping "%s": already queued', url)
                continue
            elif huey.is_locked(url):
                logger.debug('Skipping "%s": already running', url)
                continue
            logger.debug('Enqueueing "%s"', url)
            parser_cls = get_parser(os.path.basename(url))
            process(url, priority=parser_cls.PRIORITY)
            enqueued += 1
        queue_size = len([t for t in huey.pending() if t.name == 'process'])
        logger.info(
            'Enqueued %d updated files for processing. Queue size: %d',
            enqueued,
            queue_size,
        )
    return updated_files


def clean():
    expiry_intervals = {
        'weather': {
            'forecast': '3 hours',
            'current': '48 hours',
        },
        'synop': {
            'synop': '30 hours',
        },
    }
    parsed_files_expiry_intervals = {
        '%/Z__C_EDZW_%': '1 week',
    }
    with get_connection() as conn:
        with conn.cursor() as cur:
            logger.info(
                'Deleting expired weather records: %s', expiry_intervals)
            for table, table_expires in expiry_intervals.items():
                for observation_type, interval in table_expires.items():
                    cur.execute(
                        f"""
                        DELETE FROM {table} WHERE
                            source_id IN (
                                SELECT id FROM sources
                                WHERE observation_type = %s) AND
                            timestamp < current_timestamp - %s::interval;
                        """,
                        (observation_type, interval),
                    )
                    conn.commit()
                    if cur.rowcount:
                        logger.info(
                            'Deleted %d outdated %s weather records from %s',
                            cur.rowcount, observation_type, table)
                cur.execute(
                    f"""
                    UPDATE sources SET
                      first_record = record_range.first_record,
                      last_record = record_range.last_record
                    FROM (
                      SELECT
                        source_id,
                        MIN(timestamp) AS first_record,
                        MAX(timestamp) AS last_record
                      FROM {table}
                      GROUP BY source_id
                    ) AS record_range
                    WHERE sources.id = record_range.source_id;
                    """)
                conn.commit()
            logger.info(
                'Deleting expired parsed files: %s',
                parsed_files_expiry_intervals)
            for filename, interval in parsed_files_expiry_intervals.items():
                cur.execute(
                    """
                    DELETE FROM parsed_files WHERE
                        url LIKE %s AND
                        parsed_at < current_timestamp - %s::interval;
                    """,
                    (filename, interval))
                conn.commit()
                if cur.rowcount:
                    logger.info(
                        'Deleted %d outdated parsed files for pattern "%s"',
                        cur.rowcount, filename)



import datetime
import logging
import re

import dateutil.parser
import requests
from dateutil.tz import tzutc
from parsel import Selector

from brightsky.db import fetch
from brightsky.parsers import get_parser


class DWDPoller:

    urls = [
        'https://opendata.dwd.de/weather/local_forecasts/mos/MOSMIX_L/'
        'all_stations/kml/',
        'https://opendata.dwd.de/weather/local_forecasts/mos/MOSMIX_S/'
        'all_stations/kml/',
        'https://opendata.dwd.de/weather/weather_reports/synoptic/germany/'
        'json/',
        'https://opendata.dwd.de/weather/weather_reports/poi/',
    ] + [
        'https://opendata.dwd.de/climate_environment/CDC/observations_germany/'
        f'climate/hourly/{subfolder}/'
        for subfolder in [
            'air_temperature', 'cloudiness', 'dew_point', 'visibility',
            'precipitation', 'pressure', 'sun', 'wind']
    ] + [
        'https://opendata.dwd.de/climate_environment/CDC/observations_germany/'
        f'climate/10_minutes/{param}/{subfolder}/'
        for param in ['extreme_wind', 'solar']
        for subfolder in ['recent', 'historical']
    ]

    @property
    def logger(self):
        if not hasattr(self, '_logger'):
            self._logger = logging.getLogger(self.__class__.__name__)
        return self._logger

    def poll(self):
        self.logger.info("Polling for updated files")
        parsed_files = {
            row['url']: row
            for row in fetch('SELECT * FROM parsed_files')
        }
        for url in self.urls:
            for file_info in self.poll_url(url):
                if not self.matches_known_fingerprint(parsed_files, file_info):
                    yield file_info

    def poll_url(self, url):
        self.logger.debug("Loading %s", url)
        resp = requests.get(url)
        resp.raise_for_status()
        return self.parse(url, resp.text)

    def parse(self, url, resp_text):
        sel = Selector(resp_text)
        directories = []
        files = []
        for anchor_sel in sel.css('a'):
            link = anchor_sel.css('::attr(href)').extract_first()
            if link.startswith('.'):
                continue
            link_url = f'{url}{link}'
            if link.endswith('/'):
                directories.append(link_url)
            else:
                fingerprint = anchor_sel.xpath(
                    './following-sibling::text()[1]').extract_first()
                match = re.match(
                    r'\s*(\d+-\w+-\d+ \d+:\d+)\s+(\d+)', fingerprint)
                last_modified = dateutil.parser.parse(
                    match.group(1)).replace(tzinfo=tzutc())
                file_size = int(match.group(2))
                parser_cls = get_parser(link)
                if parser_cls and not parser_cls().skip_path(link):
                    files.append({
                        'url': link_url,
                        'parser': parser_cls.__name__,
                        'last_modified': last_modified,
                        'file_size': file_size,
                    })
        self.logger.debug(
            "Found %d directories and %d files at %s",
            len(directories), len(files), url)
        yield from files
        for dir_url in directories:
            yield from self.poll_url(dir_url)

    def matches_known_fingerprint(self, parsed_files, file_info):
        parsed_info = parsed_files.get(file_info['url'])
        if not parsed_info:
            return False
        last_modified_diff = abs(
            file_info['last_modified'] - parsed_info['last_modified'])
        # The downloaded file timestamp will sometimes be off from the index
        # page timestamp by one second, presumably because it is delivered from
        # a different server than the one that delivered the index page. If the
        # file was modified at 59 seconds past the minute this can lead to a
        # timestamp that is off by one minute (as seconds are not part of the
        # timestamp). To not re-process these files over and over we allow a
        # one minute time difference from the known fingerprint.
        return (
            file_info['file_size'] == parsed_info['file_size'] and
            last_modified_diff <= datetime.timedelta(minutes=1))



from setuptools import setup

import brightsky


with open('README.md') as f:
    long_description = f.read()

setup(
    name='brightsky',
    version=brightsky.__version__,
    author='Jakob de Maeyer',
    author_email='jakob@naboa.de',
    description="JSON API for DWD's open weather data.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://brightsky.dev/',
    project_urls={
        'Documentation': 'https://brightsky.dev/docs/',
        'Source': 'https://github.com/jdemaeyer/brightsky/',
        'Tracker': 'https://github.com/jdemaeyer/brightsky/issues/',
    },
    packages=['brightsky'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        'astral',
        'click',
        'coloredlogs',
        'dwdparse[lean]',
        'falcon>=3',
        'gunicorn',
        'huey',
        'parsel',
        'psycopg2-binary',
        'python-dateutil',
        'redis',
        'requests',
        'sentry-sdk',
    ],
)



import datetime
from dateutil.tz import tzutc

import pytest

from brightsky.export import DBExporter, SYNOPExporter


SOURCES = [
    {
        'observation_type': 'historical',
        'lat': 10.1,
        'lon': 20.2,
        'height': 30.3,
        'wmo_station_id': '10001',
        'dwd_station_id': 'XYZ',
        'station_name': 'Münster',
    },
    {
        'observation_type': 'historical',
        'lat': 40.4,
        'lon': 50.5,
        'height': 60.6,
        'wmo_station_id': '10002',
        'dwd_station_id': None,
        'station_name': 'Aurich',
    },
    {
        'observation_type': 'historical',
        'lat': 60.6,
        'lon': 70.7,
        'height': 80.8,
        'wmo_station_id': '10003',
        'dwd_station_id': None,
        'station_name': 'Göttingen',
    },
]
RECORDS = [
    {
        'timestamp': datetime.datetime(2020, 8, 18, 18, tzinfo=tzutc()),
        'temperature': 291.25,
        'precipitation': 0.3,
    },
    {
        'timestamp': datetime.datetime(2020, 8, 18, 19, tzinfo=tzutc()),
        'temperature': 290.25,
        'precipitation': 0.2,
    },
    {
        'timestamp': datetime.datetime(2020, 8, 18, 20, tzinfo=tzutc()),
        'temperature': 289.25,
        'precipitation': 0.1,
    },
]
FINGERPRINT = {
    'url': 'https://example.com/source.zip',
    'last_modified': datetime.datetime(2020, 8, 19, 12, 34, tzinfo=tzutc()),
    'file_size': 12345
}


@pytest.fixture
def exporter():
    exporter = DBExporter()
    exporter.export(
        [
            {**SOURCES[0], **RECORDS[0]},
            {**SOURCES[1], **RECORDS[1]},
        ],
        fingerprint=FINGERPRINT)
    return exporter


def _query_sources(db):
    return db.fetch("SELECT * FROM sources ORDER BY id")


def _query_records(db, table='weather'):
    return db.fetch(
        f"""
        SELECT * FROM {table}
        JOIN sources ON {table}.source_id = sources.id
        ORDER BY sources.id, timestamp
        """)


def test_db_exporter_creates_new_sources(db, exporter):
    db_sources = _query_sources(db)
    assert len(db_sources) == 2
    for source, row in zip(SOURCES[:2], db_sources):
        for k, v in source.items():
            assert row[k] == v


def test_db_exporter_reuses_existing_sources(db, exporter):
    exporter.export([{**SOURCES[0], **RECORDS[2]}])
    db_sources = _query_sources(db)
    assert len(db_sources) == len(SOURCES[:2])
    # Exports with only known sources should also not increase the sources_id
    # sequence
    exporter.export([{**SOURCES[2], **RECORDS[2]}])
    db_sources = _query_sources(db)
    assert db_sources[2]['id'] == db_sources[0]['id'] + 2


def test_db_exporter_creates_new_records(db, exporter):
    db_records = _query_records(db)
    for record, source, row in zip(RECORDS[:2], SOURCES[:2], db_records):
        for k, v in source.items():
            assert row[k] == v
        for k, v in record.items():
            assert row[k] == v


def test_db_exporter_updates_existing_records(db, exporter):
    record = RECORDS[0].copy()
    record['precipitation'] = 10.
    record['cloud_cover'] = 50
    exporter.export([{**SOURCES[0], **record}])
    db_records = _query_records(db)
    for k, v in record.items():
        assert db_records[0][k] == v


def test_db_exporter_updates_parsed_files(db, exporter):
    parsed_files = db.fetch("SELECT * FROM parsed_files")
    assert len(parsed_files) == 1
    for k, v in FINGERPRINT.items():
        assert parsed_files[0][k] == v


def test_db_exporter_updates_source_first_last_record(db, exporter):
    db_sources = _query_sources(db)
    assert db_sources[0]['first_record'] == RECORDS[0]['timestamp']
    assert db_sources[0]['last_record'] == RECORDS[0]['timestamp']
    exporter.export([{**SOURCES[0], **RECORDS[2]}])
    db_sources = _query_sources(db)
    assert db_sources[0]['first_record'] == RECORDS[0]['timestamp']
    assert db_sources[0]['last_record'] == RECORDS[2]['timestamp']


def test_synop_exporter(db):
    exporter = SYNOPExporter()
    assert len(_query_records(db, table='current_weather')) == 0
    # Exporter needs to merge separate records for the same source and time
    record = RECORDS[0].copy()
    record['timestamp'] = datetime.datetime.utcnow().replace(
        minute=0, second=0, microsecond=0, tzinfo=tzutc())
    extra_record = {
        'timestamp': record['timestamp'],
        'pressure_msl': 101010,
    }
    previous_record = RECORDS[1].copy()
    previous_record['timestamp'] = (
        record['timestamp'] - datetime.timedelta(minutes=30))
    exporter.export([
        {**SOURCES[0], **record},
        {**SOURCES[0], **extra_record},
        {**SOURCES[0], **previous_record},
    ])
    # Merges records for the same source and timestamp
    synop_records = _query_records(db, table='synop')
    assert len(synop_records) == 2
    assert synop_records[-1]['timestamp'] == record['timestamp']
    assert synop_records[-1]['temperature'] == record['temperature']
    assert synop_records[-1]['pressure_msl'] == extra_record['pressure_msl']
    # Updates current_weather
    # XXX: This test may be flaky as the concurrent refresh may not have
    #      finished yet. Can we somehow wait until the lock is released?
    current_weather_records = _query_records(db, table='current_weather')
    assert len(current_weather_records) == 1



import importlib
import sys
from contextlib import contextmanager

import falcon
from dateutil.tz import gettz
from dwdparse.units import convert_record, CONVERTERS
from falcon.errors import HTTPInvalidParam
from gunicorn.app.base import BaseApplication
from gunicorn.util import import_app

import brightsky
from brightsky import query
from brightsky.settings import settings
from brightsky.utils import parse_date, sunrise_sunset


@contextmanager
def convert_exceptions():
    try:
        yield
    except ValueError as e:
        raise falcon.HTTPBadRequest(description=str(e))
    except LookupError as e:
        raise falcon.HTTPNotFound(description=str(e))


class BrightskyResource:

    ALLOWED_UNITS = ['si'] + list(CONVERTERS)

    def parse_location(self, req, required=False):
        lat = req.get_param_as_float(
            'lat', required=required, min_value=-90, max_value=90)
        lon = req.get_param_as_float(
            'lon', required=required, min_value=-180, max_value=180)
        if lat != lat:
            raise HTTPInvalidParam('The value cannot be NaN', 'lat')
        elif lon != lon:
            raise HTTPInvalidParam('The value cannot be NaN', 'lon')
        return lat, lon

    def parse_source_ids(self, req):
        return (
            self._parse_list_or_single(req, 'source_id', transform=int),
            self._parse_list_or_single(req, 'dwd_station_id'),
            self._parse_list_or_single(req, 'wmo_station_id'))

    def _parse_list_or_single(self, req, name, transform=None):
        value = req.get_param_as_list(name, transform=transform)
        if value and len(value) == 1:
            return value[0]
        return value

    def parse_max_dist(self, req):
        return req.get_param_as_int(
            'max_dist', min_value=0, max_value=500000, default=50000)

    def parse_date_range(self, req):
        date_str = req.get_param('date', required=True)
        last_date_str = req.get_param('last_date')
        try:
            date = parse_date(date_str)
            if last_date_str:
                last_date = parse_date(last_date_str)
            else:
                last_date = None
        except (ValueError, OverflowError):
            raise falcon.HTTPBadRequest(
                description='Please supply dates in ISO 8601 format')
        return date, last_date

    def parse_timezone(self, req):
        tz_str = req.get_param('tz')
        if not tz_str:
            return
        tz = gettz(tz_str)
        if not tz:
            raise falcon.HTTPBadRequest(
                description='Unknown timezone: %s' % tz_str)
        return tz

    def parse_units(self, req):
        units = req.get_param('units', default='dwd').lower()
        if units not in self.ALLOWED_UNITS:
            raise falcon.HTTPBadRequest(
                description="'units' must be in %s" % (self.ALLOWED_UNITS,))
        return units

    def process_timestamp(self, row, key, timezone):
        if not row[key]:
            return
        if timezone:
            row[key] = row[key].astimezone(timezone)
        row[key] = row[key].isoformat()

    def process_sources(self, sources, timezone=None):
        for source in sources:
            self.process_timestamp(source, 'first_record', timezone)
            self.process_timestamp(source, 'last_record', timezone)


class WeatherResource(BrightskyResource):

    PRECIPITATION_FIELD = 'precipitation'
    WIND_SPEED_FIELD = 'wind_speed'

    def on_get(self, req, resp):
        date, last_date = self.parse_date_range(req)
        lat, lon = self.parse_location(req)
        source_id, dwd_station_id, wmo_station_id = self.parse_source_ids(req)
        max_dist = self.parse_max_dist(req)
        timezone = self.parse_timezone(req)
        units = self.parse_units(req)
        if timezone:
            if not date.tzinfo:
                date = date.replace(tzinfo=timezone)
            if last_date and not last_date.tzinfo:
                last_date = last_date.replace(tzinfo=timezone)
        elif date.tzinfo:
            timezone = date.tzinfo
        with convert_exceptions():
            result = self.query(
                date, last_date=last_date, lat=lat, lon=lon,
                dwd_station_id=dwd_station_id, wmo_station_id=wmo_station_id,
                source_id=source_id, max_dist=max_dist)
        self.process_sources(result['sources'])
        source_map = {s['id']: s for s in result['sources']}
        for row in result['weather']:
            self.process_row(row, units, timezone, source_map)
        resp.media = result

    def query(self, *args, **kwargs):
        return query.weather(*args, **kwargs)

    def process_row(self, row, units, timezone, source_map):
        row['icon'] = self.get_icon(row, source_map)
        if units != 'si':
            convert_record(row, units)
        self.process_timestamp(row, 'timestamp', timezone)

    def get_icon(self, row, source_map):
        if row['condition'] in (
                'fog', 'sleet', 'snow', 'hail', 'thunderstorm'):
            return row['condition']
        precipitation = row[self.PRECIPITATION_FIELD]
        wind_speed = row[self.WIND_SPEED_FIELD]
        # Don't show 'rain' icon for little precipitation, and do show 'rain'
        # icon when condition is None but there is significant precipitation
        is_rainy = (
            row['condition'] == 'rain' and precipitation is None) or (
            (precipitation or 0) > settings.ICON_RAIN_THRESHOLD)
        if is_rainy:
            return 'rain'
        elif (wind_speed or 0) > settings.ICON_WIND_THRESHOLD:
            return 'wind'
        elif (row['cloud_cover'] or 0) >= settings.ICON_CLOUDY_THRESHOLD:
            return 'cloudy'
        source = source_map[row['source_id']]
        try:
            sunrise, sunset = sunrise_sunset(
                source['lat'], source['lon'], row['timestamp'].date())
        except ValueError as e:
            daytime = 'day' if 'above' in e.args[0] else 'night'
        else:
            daytime = (
                'day' if sunrise <= row['timestamp'] <= sunset else 'night')
        if (row['cloud_cover'] or 0) >= settings.ICON_PARTLY_CLOUDY_THRESHOLD:
            return f'partly-cloudy-{daytime}'
        return f'clear-{daytime}'


class CurrentWeatherResource(WeatherResource):

    PRECIPITATION_FIELD = 'precipitation_10'
    WIND_SPEED_FIELD = 'wind_speed_10'

    def on_get(self, req, resp):
        lat, lon = self.parse_location(req)
        source_id, dwd_station_id, wmo_station_id = self.parse_source_ids(req)
        max_dist = self.parse_max_dist(req)
        timezone = self.parse_timezone(req)
        units = self.parse_units(req)
        with convert_exceptions():
            result = query.current_weather(
                lat=lat, lon=lon, dwd_station_id=dwd_station_id,
                wmo_station_id=wmo_station_id, source_id=source_id,
                max_dist=max_dist)
        self.process_sources(result['sources'])
        source_map = {s['id']: s for s in result['sources']}
        self.process_row(result['weather'], units, timezone, source_map)
        resp.media = result


class SynopResource(WeatherResource):

    PRECIPITATION_FIELD = 'precipitation_10'
    WIND_SPEED_FIELD = 'wind_speed_10'

    def query(self, *args, **kwargs):
        kwargs.pop('max_dist')
        if any(kwargs.pop(param) for param in ['lat', 'lon']):
            description = (
                "Querying by lat/lon is not supported for the synop endpoint"
            )
            raise falcon.HTTPBadRequest(description=description)
        return query.synop(*args, **kwargs)


class SourcesResource(BrightskyResource):

    def on_get(self, req, resp):
        lat, lon = self.parse_location(req)
        max_dist = self.parse_max_dist(req)
        source_id, dwd_station_id, wmo_station_id = self.parse_source_ids(req)
        with convert_exceptions():
            result = query.sources(
                lat=lat, lon=lon, dwd_station_id=dwd_station_id,
                wmo_station_id=wmo_station_id, source_id=source_id,
                max_dist=max_dist, ignore_type=True)
        self.process_sources(result.get('sources', []))
        resp.media = result


class StatusResource:

    def on_get(self, req, resp):
        resp.media = {
            'name': 'brightsky',
            'version': brightsky.__version__,
            'status': 'ok',
        }

    def on_head(self, req, resp):
        pass


def make_cors_middleware():
    cors_origins = (
        '*'
        if settings.CORS_ALLOW_ALL_ORIGINS
        else settings.CORS_ALLOWED_ORIGINS
    )
    cors_headers = (
        '*'
        if settings.CORS_ALLOW_ALL_HEADERS
        else settings.CORS_ALLOWED_HEADERS
    )
    return falcon.CORSMiddleware(
        allow_origins=cors_origins,
        expose_headers=cors_headers,
    )


def make_app():
    app = falcon.App(middleware=[make_cors_middleware()])
    app.req_options.auto_parse_qs_csv = True
    app.add_route('/', StatusResource())
    app.add_route('/weather', WeatherResource())
    app.add_route('/current_weather', CurrentWeatherResource())
    app.add_route('/synop', SynopResource())
    app.add_route('/sources', SourcesResource())
    return app


app = make_app()


class StandaloneApplication(BaseApplication):

    def __init__(self, app_uri, **options):
        self.app_uri = app_uri
        self.options = options
        super().__init__()

    def load_config(self):
        for k, v in self.options.items():
            self.cfg.set(k.lower(), v)

    def load(self):
        brightsky_mods = [
            mod for name, mod in sys.modules.items()
            if name.startswith('brightsky.')]
        for mod in brightsky_mods:
            importlib.reload(mod)
        return import_app(self.app_uri)
